#!/usr/bin/python

# XStream/py -- Stefan Alfredsson <stefan+xbox@alfredsson.org> (c) 2002

# $Id: xstream.py,v 1.2 2002/06/18 18:24:33 alfs Exp alfs $

# This code is licensed under the terms of the GNU General Public Licence
# Available from http://www.gnu.org/licenses

# Protocol details are derived from a C implementation by
# J-F Mammet <jeff@home.bidou.info>, derived in turn from
# the C# reference implementation by RUNTiME

import os,socket,sys,string

CMDLEN = 32  # length of commands in bytes

# where is the media, and what types are allowed?

# mediadirs = ["/data/kza", "/data/my_divx", "/data/dcpp"]
mediadirs = ["/tmp"]

# extensions = [".mpg", ".avi", ".mp3"]  # must be 4 chars each!
extensions = [".mpg", ".avi", "mpeg"]

# Setup the main socket for connections
def setup_server_socket(port):
    ss = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    ss.setsockopt(socket.SOL_SOCKET,socket.SO_REUSEADDR,1)
    ss.bind((socket.gethostname(), port))
    #fcntl.fcntl(ss.fileno(), FCNTL.F_SETFL, os.O_NONBLOCK)

    ss.listen(5)
    return ss

# Main function for receving messages from the client
def recv_command(cs):
    cmd = cs.recv(CMDLEN)
    print "INFO: got message '" + cmd + "' from client\n"
    return cmd

# Function to open a file -- implements some permission checks
def do_open(f):
    print "INFO: about to open file " + f
    # check permissions so it is contained within a media directory,
    # and ends in proper suffix

    for path in mediadirs:
        if string.find(f, path) == 0:
            if string.find(f, "..") == -1: # no .. escaping!!
                if f[-4:] in extensions:
                    return open(f, "rb")

    print "WARN: Refusing to open " + f + " because not in directory or not correct extension"
    raise IOError


# Function to send the media directory contents
def send_catalog(cs):
    # <SHARES><FILE><PATH>$/some/path/file.mpg</PATH></FILE>...</SHARES>
    print "INFO: about to send directory contents"
    buf = "<SHARES>\n"
    files = 0;
    for dir in mediadirs:
        print "processing ", dir
        for fn in os.listdir(dir):
            ext = fn[-4:]  # maybe strchr on "." ? i.e. handle .mpeg
            # allowed extension?
            if ext in extensions:
                # CHANGE DIR SEPARATOR TO "\\" IF ON WINDOWS!!
                buf += "<FILE><PATH>$" + dir + "/" + fn + "</PATH></FILE>\n"
                files += 1

    buf += "</SHARES>\n"
    if files:
        # found atleast one media, send listing
        send_OK(cs,len(buf))
        cs.send(buf)
    else:
        # no files found, send "error"
        send_OK(cs,0)

# sends an "OK" reply with the amount of data to read
def send_OK(cs,amount):
    msg = "%d %s" % (amount, "OK");
    msg = string.ljust(msg, CMDLEN);
    print "INFO: Sent message '" + msg + "'"
    cs.send(msg)

# sends an error reply with the message preconstructed
def send_ERR(cs,msg):
    msg = string.ljust(msg, CMDLEN);
    print "INFO: Sent message '" + msg + "'"
    cs.send(msg)

# Function for transfering data to the client
def transfer(fd, cs, amount):

    # need to store in a buffer to see how many bytes are to be sent

    buf = fd.read(amount);
    send_OK(cs, len(buf));
    cs.send(buf)

# MAIN PROGRAM LOOP -- serves commands from the client
def serve_connection(cs):

    fo = 0; # fileobject for current file

    # Protocol step 1. Send greeting

    cs.send("HELLO XBOX!")

    # Protocol step 2. Get command from other side
    alive = 1;

    while alive:
        cmd = recv_command(cs)
        if cmd != '':
            c = cmd[0:4]

            # Dispatch the various commands

            if c == "READ":
                if fo != 0:
                    transfer(fo, cs, string.atoi(cmd[5:].rstrip()))
                else:
                    print "WARN: Trying to read with no file open! " + cmd
                    send_OK(cs, 0)

            if c == "OPEN":
                if fo != 0:
                    fo.close();
                    send_OK(cs,0)
                try:
                    fo = do_open(cmd[5:].strip())
                    send_OK(cs, 0)
                except IOError, e:
                    send_ERR(cs, "-1 UNABLE TO OPEN")
                    
            if c == "*CAT":
                send_catalog(cs)

            if c == "CLSE":
                print "INFO: closing file ", fo
                try:
                    fo.close()
                except e:
                    # not functional at the moment
                    print "WARN: close failed ", e


            # Add protocol command extensions here!

            # THESE ARE EXPERIMENTAL ADDITIONS!
            
            if c == "SEEK":
                print "INFO: moving file pointer"
                fo.seek(string.atoi(cmd[5:].strip()),1)
                send_OK(cs,0)

            if c == "TELL":
                print "INFO: informing about file position"
                send_OK(cs,fo.tell())

        else:
            print "INFO: connection broken, closing down.\n"
            cs.close()
            alive = 0

            
#### main program starts here

# check if user wants extra dirs added to media dirs

args = sys.argv[1:]
for x in args:
    mediadirs.append(x)

ss = setup_server_socket(1400)

print "INFO: Waiting for commands on socket " , ss.getsockname()

while 1:
    #accept connections from outside
    (clientsocket, address) = ss.accept()

    # Check address for access control!
    # if (string.find(address[0], "192.168.") == 0):
    #if (string.find(address[0], "192.168.") == 0):
    if 1:
        print "INFO: got connection from ", address[0], " -- Serving!\n"
        # check OK, handle connection
        serve_connection(clientsocket)
    else:
        print "WARN: DENYING CONNECTION FROM ", address
        clientsocket.send("-1 You are not allowed to connect. Please contact the server admin to allow your address " + address[0] + " access (bottom of xstream.py)\n")
        clientsocket.close()
    

# EOF
